/*	Copyright (c) 1991 Richard F. Man.  All rights reserved.
 *	man@labrea.zko.dec.com, P.O. Box 6, North Chelmsford, MA 01863
 *	tabstop=4
 *
 *	$Log: msg.c%v $
 * Revision 1.1  1991/12/08  22:12:48  rfm
 * Initial revision
 *
 *
 */
#include <stdlib.h>
#include "exec.h"
#include "execpriv.h"

#define SUPPRESS_TIME	5

/*	file static queue headers
 */
static QUEUE_HDR HsemaphoreQueue, *semaphoreQueue = &HsemaphoreQueue;
static QUEUE_HDR HtimerQueue, *timerQueue = &HtimerQueue;
static QUEUE_HDR HportQueue, *portQueue = &HportQueue;
/*	this is the runnable queue
 */
static QUEUE_HDR HmsgQueue, *msgQueue = &HmsgQueue;
static QUEUE_HDR HmsgBlockQueue, *msgBlockQueue = &HmsgBlockQueue;
static QUEUE_HDR HXQueue, *XQueue = &HXQueue;

static int numberOfTimerProcesses;

static void msgDecrPortTimer(void *p);
static void msgDecrTimer(void *p);
static BOOL msgConnectPorts(STRING_ATOM sourceAtom, STRING_ATOM destinationAtom,
        CONNECTION_TYPE type);

/*	decrement the suppression timers
 */
static
void msgDecrPortTimer(void *p)
    {
    PORT_QUEUE *pQueue;

    pQueue = p;
    if (pQueue->inhibitTimer)
        --pQueue->inhibitTimer;
    if (pQueue->suppressTimer)
        --pQueue->suppressTimer;
    }

/*	decrement the thread timer
 */
static
void msgDecrTimer(void *p)
    {
    TIMER_QUEUE *pQueue;

    pQueue = p;
    if (pQueue->currentTimerInterval == 0)
        {
		if (!QueFindItem(msgQueue, pQueue->watchID))
			MsgEnQueue(pQueue->threadID, pQueue->watchID);
        pQueue->currentTimerInterval = pQueue->initialTimerInterval;
        }
    else
        --pQueue->currentTimerInterval;
    }

/*	clear a semaphore
 */
void MsgClearSemaphore(STRING_ATOM semaphoreAtom)
    {
    SEMA_QUEUE *pQueue;

    pQueue = QueFindItem(semaphoreQueue, semaphoreAtom);
    if (pQueue && !pQueue->detectOn)
        MsgEnQueue(pQueue->threadID, pQueue->watchID);
    }

/*	connect two ports normally, first suppresses the second, or first
 *	inhibits the second.
 */
static
BOOL msgConnectPorts(STRING_ATOM sourceAtom, STRING_ATOM destinationAtom, 
        CONNECTION_TYPE type)
    {
    unsigned int i;
    PORT_QUEUE *pSourceQueue, *pDestinationQueue;

    pDestinationQueue = QueFindItem(portQueue, destinationAtom);
	/*	if the ports not yet exist, create them
	 */
    if (!pDestinationQueue)
		pDestinationQueue = MsgOpenPort(destinationAtom, NULL);
    pSourceQueue = QueFindItem(portQueue, sourceAtom);
    if (!pSourceQueue)
		pSourceQueue = MsgOpenPort(sourceAtom, NULL);
	/*	find and use a free entry in the table
	 */
    for (i = 0; i < NUMBER_OF_CONNECTED_PORTS && pSourceQueue->destinations[i]; 
            ++i)
        ;
    if (i < NUMBER_OF_CONNECTED_PORTS)
        {
        pSourceQueue->destinations[i] = pDestinationQueue;
        pSourceQueue->destinationTypes[i] = type;
        return (YES);
        }
    return (NO);
    }

/*	connect an output port to an input port
 */
BOOL MsgConnectPorts(STRING_ATOM sourceAtom, STRING_ATOM destinationAtom)
    {

    return (msgConnectPorts(sourceAtom, destinationAtom, CONNECT_PORT));
    }

/*	delete a timer
 */
void MsgDeleteTimer(EXEC_ID watchID)
	{

	QueDeleteQueue(timerQueue, watchID);
	}

/*	get the first item off the runnable queue
 */
EXEC_ID MsgDeQueue(EXEC_ID *pWatchID)
    {
	MSG_QUEUE *pQueue;

    pQueue = QueDeQueue(msgQueue);
	/*	store the memory in msgBlock queue so it can be reuse fast
	 */
	QueEnQueue(msgBlockQueue, pQueue);
	if (pWatchID)
		*pWatchID = pQueue->watchID;
	return (pQueue->threadID);
    }

/*	put an item at the end of the runnable queue
 */
void MsgEnQueue(THREAD_ID threadID, EXEC_ID watchID)
    {
    MSG_QUEUE *pMsgQueue;

	/*	resue memory from the msgBlockQueue if possible
	 */
	if (!QueEmpty(msgBlockQueue))
		pMsgQueue = QueDeQueue(msgBlockQueue);
	else
		pMsgQueue = ExecAlloc(sizeof (*pMsgQueue));
	pMsgQueue->threadID = threadID;
	pMsgQueue->watchID = watchID;
	pMsgQueue->nameAtom = watchID;
	QueEnQueue(msgQueue, pMsgQueue);
    }

/*	get a message.  This performs a context switch
 */
EXEC_ID MsgGetMsg(void)
    {
    EXEC_ID watchID;
	static int clockTicks;

	/*	wait until there is something runnable
	 */
	do
		{
		++clockTicks;
		if (numberOfTimerProcesses &&
			(clockTicks % (numberOfTimerProcesses * 2)) == 0)
			{
			QueForeachQueueItem(timerQueue, msgDecrTimer);
			QueForeachQueueItem(portQueue, msgDecrPortTimer);
			}
		} while (QueEmpty(msgQueue));

    /* context switch.  When return, we have a msg
     */
    watchID = ExecSwitch();
    return (watchID);
    }

/*	source inhibits destination
 */
BOOL MsgInhibitPorts(STRING_ATOM sourceAtom, STRING_ATOM destinationAtom)
    {

    return (msgConnectPorts(sourceAtom, destinationAtom, INHIBIT_PORT));
    }

/*	put a port variable in the port queue
 */
void *MsgOpenPort(STRING_ATOM nameAtom, PORT *pPortValue)
	{
	PORT_QUEUE *pQueue;

	pQueue = ExecAlloc(sizeof (*pQueue));
	pQueue->nameAtom = nameAtom;
	pQueue->threadID = ExecGetThreadID();
	pQueue->pPortValue = pPortValue;
	QueEnQueue(portQueue, pQueue);
	return ((void *)pQueue);
	}

/*	get the current value of a port variable
 */
PORT MsgPeekPort(STRING_ATOM portAtom, PORT *pPortValue)
	{
	PORT portValue;
	PORT_QUEUE *pQueue;

    pQueue = QueFindItem(portQueue, portAtom);
	if (pQueue && pQueue->pPortValue)
		portValue = *pQueue->pPortValue;
	else
		portValue = 0;
	if (pPortValue)
		*pPortValue = portValue;
	return (portValue);
	}

/*	set a semaphore
 */
void MsgSetSemaphore(STRING_ATOM semaphoreAtom)
    {
    SEMA_QUEUE *pQueue;

    pQueue = QueFindItem(semaphoreQueue, semaphoreAtom);
    if (pQueue && pQueue->detectOn)
        MsgEnQueue(pQueue->threadID, pQueue->watchID);
    }

/*	source suppresses destination
 */
BOOL MsgSuppressPorts(STRING_ATOM sourceAtom, STRING_ATOM destinationAtom)
    {

    return (msgConnectPorts(sourceAtom, destinationAtom, SUPPRESS_PORT));
    }

/*	set up a swtch over a port variable
 */
EXEC_ID MsgWatchPort(STRING_ATOM portAtom, PORT *pPortValue)
    {
    PORT_QUEUE *pQueue;

	/*	if we are already watching this port, do nothing
	 */
    if ((pQueue = QueFindItem(portQueue, portAtom)) != NULL && 
	    pQueue->watchID)
        return (BAD_EXEC_ID);
    if (!pQueue)
        {
        pQueue = ExecAlloc(sizeof (*pQueue));
        pQueue->nameAtom = portAtom;
        pQueue->threadID = ExecGetThreadID();
        QueEnQueue(portQueue, pQueue);
        }
    pQueue->pPortValue = pPortValue;
    pQueue->watchID = ExecExecID(PORT_PREFIX);
    return (pQueue->watchID);
    }

/*	set up to watch for a semaphore
 */
EXEC_ID MsgWatchSemaphore(STRING_ATOM semaphoreAtom, BOOL detectOn)
    {
    SEMA_QUEUE *pQueue;

    if (QueFindItem(semaphoreQueue, semaphoreAtom))
        return (BAD_EXEC_ID);
    pQueue = ExecAlloc(sizeof (*pQueue));
    pQueue->detectOn = detectOn;
    pQueue->threadID = ExecGetThreadID();
    pQueue->nameAtom = semaphoreAtom;
    pQueue->watchID = ExecExecID(SEMAPHORE_PREFIX);
    QueEnQueue(semaphoreQueue, pQueue);
    return (pQueue->watchID);
    }

/*	watch a timer
 */
EXEC_ID MsgWatchTimer(unsigned int interval)
    {
    TIMER_QUEUE *pQueue;

	++numberOfTimerProcesses;
    pQueue = ExecAlloc(sizeof (*pQueue));
    pQueue->threadID = ExecGetThreadID();
    pQueue->nameAtom = pQueue->watchID = ExecExecID(TIMER_PREFIX);
    pQueue->initialTimerInterval = pQueue->currentTimerInterval = interval;
    QueEnQueue(timerQueue, pQueue);
    return (pQueue->watchID);
    }

/*	write to an output port
 */
void MsgWriteToPort(STRING_ATOM portAtom, PORT portValue)
    {
    unsigned int i;
    PORT_QUEUE *pDestination, *pQueue;

    pQueue = QueFindItem(portQueue, portAtom);
	/*	only write when noone is inhibiting us
	 */
    if (pQueue && pQueue->inhibitTimer == 0)
        {
        if (pQueue->pPortValue)
			*(pQueue->pPortValue) = portValue;
        for (i = 0; i < NUMBER_OF_CONNECTED_PORTS && 
                (pDestination = pQueue->destinations[i]) != NULL; ++i)
            {
            if (pQueue->destinationTypes[i] == INHIBIT_PORT)
                pDestination->inhibitTimer = SUPPRESS_TIME;
            else if (pQueue->destinationTypes[i] == SUPPRESS_PORT)
                pDestination->suppressTimer = SUPPRESS_TIME;
			if (pDestination->pPortValue)
				*pDestination->pPortValue = portValue;
			MsgEnQueue(pDestination->threadID, pDestination->watchID);
            }
        }
    }
